/*
** $Id: lstate.c,v 1.66 2001/07/17 17:54:46 roberto Exp $
** Global State
** See Copyright Notice in lua.h
*/


#include <stdio.h>
#include <stdlib.h>

#define LUA_PRIVATE
#include "lua.h"

#include "ldo.h"
#include "lgc.h"
#include "llex.h"
#include "lmem.h"
#include "lstate.h"
#include "lstring.h"
#include "ltable.h"
#include "ltm.h"


static void* luaHelper_ReallocFunction(void* ptr, int size, void* data, const char* allocName, unsigned int allocFlags)
{
	data, allocName, allocFlags;
	return realloc(ptr, size);
}


static void luaHelper_FreeFunction(void* ptr, void* data)
{
	free(ptr);
}


static lua_ReallocFunction luaHelper_Realloc = luaHelper_ReallocFunction;
static lua_FreeFunction luaHelper_Free = luaHelper_FreeFunction;
static void* luaHelper_memData = NULL;


void lua_getdefaultmemoryfunctions(lua_ReallocFunction* reallocFunc, lua_FreeFunction* freeFunc, void** data)
{
	*reallocFunc = luaHelper_Realloc;
	*freeFunc = luaHelper_Free;
	*data = luaHelper_memData;
}


void lua_setdefaultmemoryfunctions(lua_ReallocFunction reallocFunc, lua_FreeFunction freeFunc, void* data)
{
	if (reallocFunc)
		luaHelper_Realloc = reallocFunc;
	else
		luaHelper_Realloc = luaHelper_ReallocFunction;

	if (freeFunc)
		luaHelper_Free = freeFunc;
	else
		luaHelper_Free = luaHelper_FreeFunction;

	luaHelper_memData = data;
}


static int lua_minimumnumstrings = NUM_RESERVED + NUM_TAGS;

void lua_setminimumstringtablesize(int numstrings)
{
	if (numstrings == 0)
		lua_minimumnumstrings = NUM_RESERVED + NUM_TAGS;
	else
		lua_minimumnumstrings = numstrings;
}

static int lua_minimumglobaltableentries = 10;

void lua_setminimumglobaltablesize(int numentries)
{
	if (numentries == 0)
		lua_minimumglobaltableentries = 10;
	else
		lua_minimumglobaltableentries = numentries;
}

static int lua_minimumauxspace = 0;

void lua_setminimumauxspace(int size)
{
	lua_minimumauxspace = size;
}

struct Sopen {
  int stacksize;
  lua_State *L;
};


static void defaultFatalErrorFunc()
{
	exit(EXIT_FAILURE);
}


static void close_state (lua_State *L, lua_State *OL);


/*
** open parts that may cause memory-allocation errors
*/
static void f_luaopen (lua_State *L, void *ud) {
  int i;
  struct Sopen *so = (struct Sopen *)ud;
  if (so->stacksize == 0)
    so->stacksize = DEFAULT_STACK_SIZE;
  else
    so->stacksize += LUA_MINSTACK;
  if (so->L != NULL) {  /* shared global state? */
    L->G = G(so->L);
    L->gt = so->L->gt;  /* share table of globals */
    so->L->next->previous = L;  /* insert L into linked list */
    L->next = so->L->next;
    so->L->next = L;
    L->previous = so->L;
    luaD_init(L, so->stacksize);  /* init stack */
  }
  else {  /* create a new global state */
    global_State globalState;
    lua_State luaState;
	luaState.allocName = "global_State";
    luaState.G = &globalState;
    globalState.reallocFunc = luaHelper_Realloc;
    globalState.freeFunc = luaHelper_Free;
    globalState.memData = luaHelper_memData;
    L->G = luaM_new(&luaState, global_State);
    G(L)->reallocFunc = luaHelper_Realloc;
    G(L)->freeFunc = luaHelper_Free;
    G(L)->memData = luaHelper_memData;
    G(L)->strt.size = 0;
    G(L)->strt.nuse = 0;
    G(L)->strt.hash = NULL;
    G(L)->Mbuffer = NULL;
    G(L)->Mbuffsize = 0;
    G(L)->rootproto = NULL;
    G(L)->rootcl = NULL;
    G(L)->roottable = NULL;
    G(L)->rootudata = NULL;
    G(L)->TMtable = NULL;
    G(L)->sizeTM = 0;
    G(L)->ntag = 0;
    G(L)->nblocks = sizeof(lua_State) + sizeof(global_State);
	/* build methods tables */
	for (i = 0; i < LUA_NTYPES; i++)
		G(L)->methods[i] = luaH_new(L, 0);
	for (i = 0; i <= LUA_TPOINTER; i++)
		G(L)->methods[i]->methods = L->G->methods[LUA_TTABLE];
    luaD_init(L, so->stacksize);  /* init stack */
    luaM_setname(L, "Lua_GlobalsHash");
    L->gt = luaH_new(L, lua_minimumglobaltableentries);  /* table of globals */
    luaM_setname(L, "Lua_Type2TagHash");
    G(L)->type2tag = luaH_new(L, 10);
    luaM_setname(L, "Lua_Registry");
    G(L)->registry = luaH_new(L, 0);
    luaM_setname(L, "Lua_WeakRegistry");
    G(L)->weakregistry = luaH_new(L, 0);
    /* make weakregistry weak */
    G(L)->weakregistry->weakmode = LUA_WEAK_KEY | LUA_WEAK_VALUE;
//    luaS_resize(L, MINPOWER2);
    G(L)->minimumstrings = lua_minimumnumstrings;
	luaS_resize(L, G(L)->minimumstrings);
    luaX_init(L);
    luaT_init(L);
    G(L)->GCthreshold = 4*G(L)->nblocks;
	G(L)->fatalErrorFunc = defaultFatalErrorFunc;
    G(L)->lockData = NULL;
    G(L)->lockFunc = NULL;
    G(L)->unlockFunc = NULL;
	luaO_openspace(L, lua_minimumauxspace, l_char);
  }
}


LUA_API void lua_setlockfunctions(lua_State *L, void (*lockFunc)(void *),
                                  void (*unlockFunc)(void *), void *lockData)
{
  L->G->lockFunc = lockFunc;
  L->G->unlockFunc = unlockFunc;
  L->G->lockData = lockData;
}


LUA_API void lua_setfatalerrorfunction(lua_State *L, void (*fatalErrorFunc)())
{
	L->G->fatalErrorFunc = fatalErrorFunc;
}


LUA_API lua_State *lua_newthread (lua_State *OL, int stacksize) {
  struct Sopen so;
  lua_State *L;
  if (OL)
  {
	  lua_lock(OL);
  L = luaM_new(OL, lua_State);
  }
  else
  {
	  global_State globalState;
	  lua_State luaState;
	  luaState.allocName = "lua_State";
	  luaState.G = &globalState;
	  globalState.reallocFunc = luaHelper_Realloc;
	  globalState.freeFunc = luaHelper_Free;
	  globalState.memData = luaHelper_memData;
	  L = luaM_new(&luaState, lua_State);
  }
  if (L) {  /* allocation OK? */
    L->G = NULL;
    L->stack = NULL;
    L->stacksize = 0;
    L->ci = &L->basefunc;
    L->basefunc.prev = NULL;
    L->errorJmp = NULL;
    L->callhook = NULL;
    L->linehook = NULL;
    L->allowhooks = 1;
	L->allocName = NULL;
    L->next = L->previous = L;
	L->allocName = NULL;
	L->allocFlags = 0;
	L->mainFunctionAllocFlags = 0;
    so.stacksize = stacksize;
    so.L = OL;
    if (luaD_runprotected(L, f_luaopen, &so) != 0) {
      /* memory allocation error: free partial state */
      close_state(L, OL);
      L = NULL;
    }
  }
  if (OL) lua_unlock(OL);
  return L;
}


static void close_state (lua_State *L, lua_State *OL) {
  global_State globalState;
  lua_State luaState;
  luaState.G = &globalState;
  globalState.reallocFunc = G(L)->reallocFunc;
  globalState.freeFunc = G(L)->freeFunc;
  globalState.memData = G(L)->memData;

  if (OL != NULL) {  /* are there other threads? */
    lua_assert(L->previous != L);
    L->previous->next = L->next;
    L->next->previous = L->previous;
  }
  else if (G(L)) {  /* last thread; close global state */
    luaC_callallgcTM(L);  /* call GC tag methods for all udata */
    luaC_collect(L, 1);  /* collect all elements */
    lua_assert(G(L)->rootproto == NULL);
    lua_assert(G(L)->rootudata == NULL);
    lua_assert(G(L)->rootcl == NULL);
    lua_assert(G(L)->roottable == NULL);
    luaS_freeall(L);
    luaM_freearray(L, G(L)->TMtable, G(L)->sizeTM, struct TM);
    luaM_freearray(L, G(L)->Mbuffer, G(L)->Mbuffsize, l_char);
    luaM_freelem(&luaState, L->G, global_State);
  }
  if (OL)
  {
  luaM_freearray(OL, L->stack, L->stacksize, TObject);
  luaM_freelem(OL, L, lua_State);
  }
  else
  {
    luaM_freearray(&luaState, L->stack, L->stacksize, TObject);
    luaM_freelem(&luaState, L, lua_State);
  }
}

LUA_API void lua_close (lua_State *L) {
  lua_State *OL;
  lua_assert(L != lua_state || lua_gettop(L) == 0);
  lua_lock(L);
  OL = L->next;  /* any surviving thread (if there is one) */
  if (OL == L) OL = NULL;  /* no surviving threads */
  close_state(L, OL);
  if (OL) lua_unlock(OL);  /* cannot unlock over a freed state */
  lua_assert(L != lua_state || memdebug_numblocks == 0);
  lua_assert(L != lua_state || memdebug_total == 0);
}

