///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaState (Lua Wrapper) source distribution
// and is Copyright 2000-2001 by Joshua C. Jensen (jjensen@workspacewhiz.com).
// The latest version may be obtained from http://www.workspacewhiz.com/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#include <windows.h>
#endif WIN32

#include <io.h>
#include "LuaState.h"
#include "LuaStateFunctions.h"

struct FindFileInfo
{
	FindFileInfo() : m_handle(NULL) {}
	~FindFileInfo() {  delete m_handle;  }
	HANDLE m_handle;
	WIN32_FIND_DATA m_fd;
};


static int LS_FindFirstFile(LuaState state)
{
	LuaObject wildcardObj(state, 1);
	LPCSTR wildcard = wildcardObj.GetString();

	FindFileInfo* info = new FindFileInfo;
	info->m_handle = FindFirstFile(wildcard, &info->m_fd);
	if (info->m_handle == INVALID_HANDLE_VALUE)
	{
		delete info;
		return 0;
	}

	state.NewUserDataBox(info);
	state.GetGlobal("FileFind");
	state.SetMethods(-2);
	return 1;
}


static int LS_FindNextFile(LuaState state)
{
	LuaObject infoObj(state, 1);
	if (!infoObj.IsUserData())
		return 0;
	FindFileInfo* info = static_cast<FindFileInfo*>(infoObj.GetUserData());
	BOOL ret = FindNextFile(info->m_handle, &info->m_fd);
	if (!ret)
	{
		FindClose(info->m_handle);
		info->m_handle = NULL;
		return 0;
	}

	state.PushBool(true);
	return 1;
}


static int LS_FindClose(LuaState state)
{
	LuaObject infoObj(state, 1);
	if (!infoObj.IsUserData())
		return 0;
	FindFileInfo* info = static_cast<FindFileInfo*>(infoObj.GetUserData());
	delete info;
	return 0;
}


static int LS_FindGetFileName(LuaState state)
{
	LuaObject infoObj(state, 1);
	if (!infoObj.IsUserData())
		return 0;
	FindFileInfo* info = static_cast<FindFileInfo*>(infoObj.GetUserData());
	state.PushString(info->m_fd.cFileName);
	return 1;
}


static int LS_FindIsDirectory(LuaState state)
{
	LuaObject infoObj(state, 1);
	if (!infoObj.IsUserData())
		return 0;
	FindFileInfo* info = static_cast<FindFileInfo*>(infoObj.GetUserData());
	state.PushBool((info->m_fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0);
	return 1;
}

static int LS_GetFileSize(LuaState state)
{
	LuaObject fileNameObj(state, 1);
	if (!fileNameObj.IsString())
		return 0;
	WIN32_FIND_DATA fd;
	HANDLE handle = FindFirstFile(fileNameObj.GetString(), &fd);
	state.PushNumber(fd.nFileSizeLow);
	FindClose(handle);
	return 1;
}

static int LS_GetFileWriteTime(LuaState state)
{
	LuaObject fileNameObj(state, 1);
	if (!fileNameObj.IsString())
		return 0;
	WIN32_FIND_DATA fd;
	HANDLE handle = FindFirstFile(fileNameObj.GetString(), &fd);
	state.PushNumber(fd.ftLastWriteTime.dwLowDateTime);
	state.PushNumber(fd.ftLastWriteTime.dwHighDateTime);
	FindClose(handle);
	return 2;
}

static int LS_SetFileWriteTime(LuaState state)
{
	LuaObject fileNameObj(state, 1);
	if (!fileNameObj.IsString())
		return 0;
	FILETIME createTime;
	FILETIME accessTime;
	FILETIME writeTime;
	LPCSTR fileName = fileNameObj.GetString();
	HANDLE handle = CreateFile(fileName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	GetFileTime(handle, &createTime, &accessTime, &writeTime);
	writeTime.dwLowDateTime = state.ToInteger(2);
	writeTime.dwHighDateTime = state.ToInteger(3);
	SetFileTime(handle, &createTime, &accessTime, &writeTime);
	CloseHandle(handle);
	return 0;
}

static int LS_access(LuaState state)
{
	LuaObject fileNameObj(state, 1);
	if (!fileNameObj.IsString())
		return 0;
	LuaObject typeObj(state, 2);
	LPCSTR type = NULL;
	if (typeObj.IsString())
		type = typeObj.GetString();

	int access = 0;
	if (type)
	{
		if (strcmp(type, "w") == 0)
			access = 2;
		else if (strcmp(type, "r") == 0)
			access = 4;
		else if (strcmp(type, "rw") == 0)
			access = 6;
	}

	state.PushNumber(_access(fileNameObj.GetString(), access));
	return 1;
}



static int LS_NewTableSize(LuaState state)
{
	lua_newtablesize(state, (int)lua_tonumber(state, 1));
	return 1;
}


static const LuaFunctionList s_fileFindFunctions[] =
{
	{ "First", LS_FindFirstFile },
	{ "Next", LS_FindNextFile },
	{ "Close", LS_FindClose },
	{ "GetFileName", LS_FindGetFileName },
	{ "IsDirectory", LS_FindIsDirectory },
	{ NULL, NULL },
};


static const LuaFunctionList s_fileFunctions[] =
{
	{ "GetSize",		LS_GetFileSize },
	{ "GetWriteTime",	LS_GetFileWriteTime },
	{ "SetWriteTime",	LS_SetFileWriteTime },
	{ "access",			LS_access },
	{ NULL, NULL },
};


void ScriptFunctionsRegister(lua_State* L)
{
	LuaState state(L);
	state.Register("FindFirstFile",		LS_FindFirstFile);
	state.Register("FindNextFile",		LS_FindNextFile);
	state.Register("FindClose",			LS_FindClose);
	state.Register("FindGetFileName",	LS_FindGetFileName);
	state.Register("FindIsDirectory",	LS_FindIsDirectory);
	state.Register("GetFileSize",		LS_GetFileSize);
	state.Register("GetFileWriteTime",	LS_GetFileWriteTime);
	state.Register("SetFileWriteTime",	LS_SetFileWriteTime);
	state.Register("newtablesize",		LS_NewTableSize);
	state.Register("access",			LS_access);
#include "HelperLuaFunctions.h"
	LuaAutoBlock autoBlock(state);
	
	LuaObject fileObj = state.GetGlobal("File");
	fileObj.Register(s_fileFunctions);
	
	LuaObject fileFindObj = state.GetGlobals().CreateTable("FileFind");
	fileFindObj.Register(s_fileFindFunctions);
}


