#include "LuaState.h"

/**
	The LuaState class is able to make Lua callbacks look like natural extensions
	of the LuaState class.
**/
static int LS_PrintNumber(LuaState state)
{
	// Retrieve the number passed on the stack.
	LuaObject numberObj(state, 1);

	// Verify it is a number and print it.
	if (numberObj.IsNumber())
		printf("%f\n", numberObj.GetNumber());

	// No return values.
	return 0;
}


/**																		  
	It is still possible to use LuaState with a Lua-style callback.
**/
static int LS_Add(lua_State* L)
{
	// Using a special LuaState constructor, we can wrap the lua_State pointer
	// very conveniently.
	LuaState state(L);

	// Retrieve the numbers passed on the stack.
	LuaObject number1Obj(state, 1);
	LuaObject number2Obj(state, 2);

	// Verify it is a number and print it.
	if (number1Obj.IsNumber()  &&  number2Obj.IsNumber())
	{
		state.PushNumber(number1Obj.GetNumber() + number2Obj.GetNumber());
	}
	else
	{
		state.PushNumber(0.0);
	}

	// 1 return value.
	return 1;
}


/**
	Demonstrate the use of the simple .ini script-like functions for writing
	values to a script.
**/
void DoScriptIniWriteTest()
{
	LuaStateOwner state(false);

	state.ConfigSetReal("Environment", "WindSpeed", 55.0);
	state.ConfigSetString("Environment", "TypeStr", "Overcast");
	state.ConfigSetInteger("AnotherGroup", "IntegerValue", 1);

	state.WriteLuaGlobalsFile("ScriptIniTest.dmp");
}


/**
	Demonstrate the use of the simple .ini script-like functions for reading
	values from a script.
**/
void DoScriptIniReadTest()
{
	LuaStateOwner state;

	state.DoFile("ScriptIniTest.dmp");

	float windSpeed = state.ConfigGetReal("Environment", "WindSpeed", 0.0);
	const char* typeStr = state.ConfigGetString("Environment", "TypeStr", "Clear");
	int integerValue = state.ConfigGetInteger("AnotherGroup", "IntegerValue");
}


/**
	Demonstrate registering callback functions for the Lua script.
**/
void DoScriptCallbackTest()
{
	LuaStateOwner state;

	state.Register("PrintNumber", LS_PrintNumber);
	state.Register("Add", LS_Add);

	state.DoFile("ScriptCallbackTest.lua");
}


/**
	Demonstrate reading and saving a script.
**/
void DoScriptSaveTest()
{
	LuaStateOwner state;

	state.DoFile("ScriptSaveTest.lua");
	state.WriteLuaGlobalsFile("ScriptSaveTest.dmp");
}


/**
	Demonstrates walking an array table.
**/
void DoScriptArrayTest()
{
	LuaStateOwner state;
	state.DoFile("ScriptArrayTest.lua");
	LuaObject testTableObj = state.GetGlobals().GetByName("TestArray");
	// or							state.GetGlobal("TestArray");
	for (int i = 1; ; ++i)
	{
		LuaAutoBlock block(state);	// Automatic stack fixups.
		LuaObject entryObj = testTableObj.GetByIndex(i);
		if (entryObj.IsNil())
			break;
		if (entryObj.IsNumber())		// IsNumber() must ALWAYS come first.
			printf("%f\n", entryObj.GetNumber());
		else if (entryObj.IsString())	// IsString() returns true for a number or string.
			printf("%s\n", entryObj.GetString());
	}
}


static int LS_PointerCall(LuaState state)
{
	bool isPointer = state.IsPointer(1);
	const void* ptr = state.ToPointer(1);
	return 0;
}


void TestPointer()
{
	LuaStateOwner state;
	state.Register("PointerCall", LS_PointerCall);
	state.GetGlobal("PointerCall");
	state.PushPointer((const void*)0xfedcba98);
	state.Call(1, 0);
}


int __cdecl main(int argc, char* argv[])
{
	DoScriptIniWriteTest();
	DoScriptIniReadTest();
	DoScriptCallbackTest();
	DoScriptSaveTest();
	DoScriptArrayTest();
	TestPointer();

	return 0;
}
